/* 2005/03/23 */

/* A Link-layer indications program for wireless devices
   using RTNETLINK. */

/* PURPOSE:
	An IPv6 Router Solicitation is transmitted when a 
	host receives an association (sorry, no RSN) from
	802.11 wireless LAN..

	This is a hacked example. Code style has been ignored

	Its primary uses are in my experiments and for
	illustration purposes.

*/

/*************************************************************/
/*
Please note that either this license MUST be displayed
or a particular instance of this code may be converted to 
the GNU General Public License v2 or greater.

Please see README.txt for details. 
*/

/*
Copyright (c) 2005, Monash University
All rights reserved.
Contributors: Greg Daley

Redistribution and use in source and binary forms, with or
without modification, are permitted provided that the following
conditions are met:

* Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

* Redistributions in binary form must reproduce the above
  copyright notice, this list of conditions and the following
  disclaimer in the documentation and/or other materials
  provided with the distribution.

* Neither the name of Monash University nor the names of its
  contributors may be used to endorse or promote products derived
  from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY
WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.

*/
/*************************************************************/




/*  GD: Note to self: 20040918 origin
    GD: Note to self: 20050110 evtest

     REF RFC 3549, for RTNETLINK msg  formats and types.


*/

/**************************************************************************/

/*ipv6 socket headers  AAA */
#include <sys/socket.h>
#include <netinet/in.h>
/* interface request headers  BBB */
#include <sys/ioctl.h>
#include <net/if.h>
/* extra ifreq headers*/
#include <string.h>

/* rtnetlink socket includes */
#include <asm/types.h>
#include <linux/netlink.h>
#include <linux/rtnetlink.h>
#include <sys/socket.h>


/* local diagnostic headers */
#include <stdio.h>
#include <stdlib.h>		/* exit(1)... */
#include <errno.h>

/* time information */
#include <time.h>
#include <sys/time.h>


#ifndef IFLA_WIRELESS
#define IFLA_WIRELESS 11	/* ???? */
#endif				/*IFLA_WIRELESS */

#ifndef _LINUX_IF_H
#define _LINUX_IF_H
#define _UNDEF_LINUX_IF_H
#endif				/*  _LINUX_IF_H */

#include <linux/wireless.h>

#ifdef _UNDEF_LINUX_IF_H
#undef _UNDEF_LINUX_IF_H
#undef _LINUX_IF_H
#endif				/* _UNDEF_LINUX_IF_H */


#include "fdarray.h"
#include "inputs.h"
#include "eventproc.h"


extern int bind_rtsock(uint32_t groups);
extern int open_v6sock(void);
extern int open_icmpv6receiver(void);

extern int get_ifindex(int sock, const char *name);

extern int get_mac_rtnetlink(int sock, int ifindex, unsigned char *mac);
extern int get_linklocal_rtnetlink(int sock, int ifindex, unsigned char *lladdr);

/**************************************************************************/


int main(int argc, char *argv[])
{
	char name[IFNAMSIZ + 1];

	int ifindex;
	uint8_t macaddr[6];
	/*uint8_t prefix[16];*/
	int macaddrset = 0;
	struct in6_addr *src = NULL;
	struct in6_addr srcbuff;
	int usesrcaddr = 0;
	int tentative = 0;
	int cisco=0;
	int landmark=0;
	struct link_status state;
	int v6sock;
	int fdarray[SOCKET_INDEX_MAX+1]; 
	int (*funcarray[SOCKET_INDEX_MAX+1])(int, struct link_status *,
							struct prefixlist *,
							struct prefixlist *);

    /***********************************************************************/
	
	memset(macaddr, 0, 6 * sizeof(uint8_t));
	/*memset(prefix, 0, sizeof(struct in6_addr));*/
	memset(&srcbuff, 0, sizeof(struct in6_addr));

	
	if ((argc > 1) && !strncmp("-C", argv[argc -1], 3)) {
		printf("found -C\n");
	 	cisco=1;
                argc--;
        }
	if ((argc > 1) && !strncmp("-L", argv[argc -1], 3)) {
		printf("found -L\n");
	 	landmark=1;
                argc--;
        }
	if ((argc > 1) && (argv[argc - 1][0] != '-')) {
		sscanf(argv[argc - 1],
		       "%02hhx:%02hhx:%02hhx:%02hhx:%02hhx:%02hhx",
		       &macaddr[0], &macaddr[1], &macaddr[2], &macaddr[3],
		       &macaddr[4], &macaddr[5]);
		macaddrset = 1;
		argc--;
	}
	if ((argc > 1) && !strncmp("-t", argv[ argc-1], 3)) {
		printf("found -t\n");
		tentative = 1;
		argc--;
	}

	if ((argc > 1) && !strncmp("-s", argv[argc -1], 3)) {
		printf("found -s\n");
		usesrcaddr = 1;
		argc--;
	}

	printf("argc %d\n", argc);
	if (argc != 2) {
		fprintf(stderr,
			"Usage: %s <interface> [-s  [-t] [<ethmac>] [-L] [-C]\n"
			"\t-s:  Use unicast source address\n"
			"\t-t:  Use tentative link-layer addresses\n"
			"\t-L:  Use prefix Landmarking (GD:always on)\n"
			"\t-C:  Cisco Aironet (airo) driver in use\n"
			"\t<ethmac>:  XX:XX:XX:XX:XX:XX\n", argv[0]);
		exit(1);
	}
	strncpy(name, argv[1], IFNAMSIZ);




    /***********************************************************************/

	if ((fdarray[SOCKET_INDEX_NTLINK] = bind_rtsock(RTMGRP_LINK)) < 0) {
		fprintf(stderr, "unable to bind netlink socket\n");
		exit(1);
	}

	if ((v6sock = open_v6sock()) < 0) {
		fprintf(stderr, "unable to get ipv6 socket\n");
		exit(1);
	}

	if ((ifindex = get_ifindex(v6sock, name)) < 0) {
		fprintf(stderr, "unable to get ifindex\n");
		exit(1);
	}
	if (!macaddrset && get_mac_rtnetlink(fdarray[SOCKET_INDEX_NTLINK],
						ifindex, macaddr) < 0) {

		printf("mac address setup failed on %s\n", name);
		return -1;
	}
	printf ("MAC is %2.2hhx:%2.2hhx:%2.2hhx:%2.2hhx:%2.2hhx:%2hhx\n",
	     macaddr[0], macaddr[1], macaddr[2], macaddr[3], macaddr[4],
	     macaddr[5]);

	if (usesrcaddr) {
		src = &srcbuff;
		if (get_linklocal_rtnetlink(fdarray[SOCKET_INDEX_NTLINK],
					 ifindex, (unsigned char *) src) <= 0){
			printf("lladdr setup failed %s\n", name);
			return -1;
		}
		printf("Link-local address: "
			"%hx:%hx:%hx:%hx:%hx:%hx:%hx:%hx\n",
			ntohs(src->s6_addr16[0]),
			ntohs(src->s6_addr16[1]),
			ntohs(src->s6_addr16[2]),
			ntohs(src->s6_addr16[3]),
			ntohs(src->s6_addr16[4]),
			ntohs(src->s6_addr16[5]),
			ntohs(src->s6_addr16[6]),
			ntohs(src->s6_addr16[7]));
	}
	if ((fdarray[SOCKET_INDEX_ICMPV6] = open_icmpv6receiver()) < 0) {
                fprintf(stderr, "unable to get ipv6 socket\n");
                exit(1);
        }

    /***********************************************************************/


	funcarray[SOCKET_INDEX_V6SOCK]=empty_fn;
	funcarray[SOCKET_INDEX_NTLINK]=get_all_events; 
	funcarray[SOCKET_INDEX_ICMPV6]=process_ra;


    /***********************************************************************/
	memset(&state, 0 , sizeof(state));
	memcpy(&state.source, &srcbuff, sizeof(struct in6_addr));
	/*memcpy(&state.landmark, &prefix, sizeof(struct in6_addr));*/
	state.landplen=128;
	state.ifindex=ifindex;
	state.tentative=tentative;
	state.cisco=cisco;
	state.v6sock=v6sock;
	state.rasock=fdarray[SOCKET_INDEX_ICMPV6];
	memcpy(state.macaddr, macaddr, 6);
    /***********************************************************************/


	process_events(fdarray, &state,funcarray);
	return 0;
}




