
/* 2005/07/02 */

/* An event processing subsystems for devices which perform DNA */ 

/* PURPOSE:
	An IPv6 Router Solicitation is transmitted when a 
	host receives an association (sorry, no RSN) from
	802.11 wireless LAN..

	This is a hacked example. Code style has been ignored

	Its primary uses are in my experiments and for
	illustration purposes.
*/

/*************************************************************/
/*
Please note that either this license MUST be displayed
or a particular instance of this code may be converted to 
the GNU General Public License v2 or greater.

Please see README.txt for details. 
*/

/*
Copyright (c) 2005, Monash University
All rights reserved.
Contributors: Greg Daley

Redistribution and use in source and binary forms, with or
without modification, are permitted provided that the following
conditions are met:

* Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

* Redistributions in binary form must reproduce the above
  copyright notice, this list of conditions and the following
  disclaimer in the documentation and/or other materials
  provided with the distribution.

* Neither the name of Monash University nor the names of its
  contributors may be used to endorse or promote products derived
  from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY
WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.

*/
/*************************************************************/




/*  GD: Note to self: 20040918 origin
    GD: Note to self: 20050110 evtest
    GD: Note to self: 20050701 evtestrsmatch

     REF RFC 3549, for RTNETLINK msg  formats and types.


*/

/**************************************************************************/

/*ipv6 socket headers  AAA */
#include <netinet/in.h>
#include <sys/socket.h>

/* memcmp/memcpy*/
#include <string.h>

/* rtnetlink socket includes */
#include <sys/types.h>
#include <asm/types.h>
#include <linux/netlink.h>
#include <linux/rtnetlink.h>
#include <sys/socket.h>


/* local diagnostic headers */
#include <stdio.h>
#include <stdlib.h>		/* exit(1)... */
#include <errno.h>

/* uint8_t */
#include <stdint.h>

/* time information */
#include <time.h>
#include <sys/time.h>


#ifndef IFLA_WIRELESS
#define IFLA_WIRELESS 11	/* ???? */
#endif				/*IFLA_WIRELESS */

#ifndef _LINUX_IF_H
#define _LINUX_IF_H
#define _UNDEF_LINUX_IF_H
#endif				/*  _LINUX_IF_H */

#include <net/if.h>
#include <linux/wireless.h>

#ifdef _UNDEF_LINUX_IF_H
#undef _UNDEF_LINUX_IF_H
#undef _LINUX_IF_H
#endif				/* _UNDEF_LINUX_IF_H */


/* select() system call */
#include <sys/select.h>


#include "fdarray.h"
#include "prefixlist.h"
#include "inputs.h"
#include "eventproc.h"
#include "dna.h"


extern int send_rs(int sock, int ifindex, uint8_t macaddr[], int tentative,
		   struct in6_addr *src, struct in6_addr *dst, uint8_t *
		   prefix);


int get_li_rtnetlink(int sock, struct link_status *state);

int get_all_events(int fd, struct link_status *state, struct prefixlist
                        *linkplist, struct prefixlist *recvplist);
void timerrestore(struct link_status *state, struct prefixlist *linkplist,
				struct prefixlist *recvplist);




/**************************************************************************/



void process_events(int fdarray[], struct link_status *state,
		 	int (*funcarray[])(int, struct link_status *,
						 struct prefixlist *,
						 struct prefixlist *)){
	int i, rval; 
	int maxsock=-1;
	int change=0;
	int listenstart, listenend;
	struct prefixlist recvplist, linkplist;
	fd_set rfds, efds;

	initprefixlist(&recvplist);
	initprefixlist(&linkplist);
	linkplist.complete=1;

	state->tv.tv_sec=4*RTR_SOLICITATION_INTERVAL; state->tv.tv_usec=0;

	listenstart=SOCKET_INDEX_NTLINK;
	listenend=SOCKET_INDEX_MAX;

	FD_ZERO(&efds);
	for(i=listenstart; i <= listenend; i++){
		if(fdarray[i]> maxsock){
			maxsock=fdarray[i];
		}
		FD_SET(fdarray[i], &efds);
	}
	
	
	while (1) {
		fprintf(stderr, "entering event loop, waiting\n");

		change=0;

		FD_ZERO(&rfds);
		for(i=listenstart; i <= listenend; i++){
			FD_SET(fdarray[i], &rfds);
		}

		if((rval = select(maxsock+1, &rfds, NULL, &efds, 
				&state->tv))<0){
			break;
		}

		printf("link prefixlist ageing %p\n",&linkplist);	
		age_prefixlist(&linkplist);
		printf("received prefixlist ageing %p\n",&recvplist);	
		age_prefixlist(&recvplist);

		if(!rval){
			timerrestore(state, &linkplist, &recvplist);
			continue;
		}

		fprintf(stderr,"exited select %d (%lu.%6.6lu)\n", rval,
			state->tv.tv_sec, state->tv.tv_usec);

		for(i=listenstart; i <= listenend; i++){
			if(FD_ISSET(fdarray[i], &efds)){
				break;
			}
		}

		for(i=listenstart; i <= listenend; i++){		
			if(!FD_ISSET(fdarray[i], &rfds)){
				continue;
			}
			rval = funcarray[i](fdarray[i], state, &linkplist,
						&recvplist);
			if(rval){
				change=changeinference(state, &linkplist, &recvplist);
			}
		}
		if(change){	
			printf("----------------------------------\n");
			printf("Link Change!\n");
			print_prefixlist(&linkplist);
			printf("----------------------------------\n");
		}
	}
}



int process_link_event(struct link_status *state, struct nlmsghdr *nlm)
{
	struct ifinfomsg *infom;
	int ml;
	unsigned int rtal;
	struct rtattr *rta;
	unsigned char *rtap;	/* RTA_PAYLOAD */
	struct sockaddr *sll;
	static int saved_wireless=0;

	struct iw_event *iwe;
	int rval = 0;
	int wireless_event=0;
	unsigned char noaddr[8];

	memset(&noaddr, 0, 8);

	infom = NLMSG_DATA(nlm);
	fprintf(stderr,"nll: %d nlf:%x ifif:%d ifit:%d ifind:%d flags: %x change %x \n",
		nlm->nlmsg_len, nlm->nlmsg_flags,
		infom->ifi_family, infom->ifi_type,
		infom->ifi_index, infom->ifi_flags, infom->ifi_change); 
	ml = nlm->nlmsg_len;
	rta = IFLA_RTA(infom);

	while (RTA_OK(rta, ml)) {
		/*  fprintf(stderr,"rok:%p  rtal: %d  ml%d ",
		   rta, rta->rta_len, rta->rta_type); */

		rtap = RTA_DATA(rta);
		rtal = RTA_PAYLOAD(rta);
		switch (rta->rta_type) {
		case IFLA_WIRELESS:
			iwe = (struct iw_event *) rtap;
			fprintf(stderr,"WIRELESS %hx (l:%d) ", iwe->cmd, iwe->len);
			switch (iwe->cmd) {
			case SIOCGIWAP:
				/* always family 1 (sockaddr) */
				sll = &iwe->u.ap_addr;
				wireless_event=1;
				fprintf(stderr,"SIOCGIWAP ");
				if (!memcmp(&sll->sa_data, noaddr, 6)) {
					/* disassociate */
					fprintf(stderr,"disassoc ");
				} else {
					fprintf(stderr,"associate "
					       "%.2hhx:%.2hhx:"
					       "%.2hhx:%.2hhx:"
					       "%.2hhx:%.2hhx",
					       sll->sa_data[0],
					       sll->sa_data[1],
					       sll->sa_data[2],
					       sll->sa_data[3],
					       sll->sa_data[4],
					       sll->sa_data[5]);
					rval = infom->ifi_index;
				}
				break;
			case SIOCSIWESSID:
				fprintf(stderr," SIOCSIWESSID");
				break;
			default:
				break;
			}
			break;
		case IFLA_BROADCAST:
			fprintf(stderr,"type IFLA_BROADCAST:%d ",rta->rta_type);
			break;
		case IFLA_IFNAME:
			fprintf(stderr,"type IFLA_IFNAME:%d (%s) ",
				rta->rta_type, rtap);	
			rval = infom->ifi_index;
			break;
		case IFLA_MTU:
			fprintf(stderr,"type IFLA_MTU:%d ",rta->rta_type);
			break;
		case IFLA_QDISC:
			fprintf(stderr,"type IFLA_QDISC:%d (%s)", rta->rta_type,
				rtap);	
			break;
		case IFLA_STATS:
			fprintf(stderr,"type IFLA_STATS:%d ", rta->rta_type);	
			break;
#if 0
		case IFLA_TXQLEN:
			fprintf(stderr,"type IFLA_TXQLEN:%d ", rta->rta_type);	
			break;
		case IFLA_MAP:
			fprintf(stderr,"type IFLA_MAP:%d ", rta->rta_type);	
			break;
		case IFLA_WEIGHT:
			fprintf(stderr,"type IFLA_WEIGHT:%d ", rta->rta_type);	
			break;
#endif /* 0 */
		default:
			fprintf(stderr,"type:%d ", rta->rta_type);
			break;
		}
		rta = RTA_NEXT(rta, ml);
	}
	if(rval && wireless_event){
		if(! ((((infom->ifi_flags & IFF_RUNNING)) && 
				(infom->ifi_change == 1)) || state->cisco) ){
                	fprintf(stderr,"Not a running wireless interface(EXP)!\n");
			saved_wireless=1;
                	return 0;
		}
	}

        if(rval && ! wireless_event){
		if((state->recvdlinkup)&& (rval==state->ifindex)&&
			(infom->ifi_flags & IFF_RUNNING)){
			
                	fprintf(stderr,"Now is a running wireless interface(EXP)!\n");
			state->recvdlinkup=0;
			return rval;
		}
		if(!((infom->ifi_flags & IFF_RUNNING)&&
					(infom->ifi_change == 1))){
                	fprintf(stderr,"Not a link up indication (EXP)!\n");
                	return 0;
		}
        }


	return rval;
}

int get_li_rtnetlink(int rtnetlink_socket, struct link_status *state)
{
	int rval = 0;
	int recvsize;
	struct nlmsghdr *nlm;
	struct {
		struct nlmsghdr nh;
		char attrbuf[1024];
	} req;

	memset(&req, 0, sizeof(req));

	fprintf(stderr, "Collecting event from rtnetlink socket %d\n",
			 rtnetlink_socket);

	if ((recvsize = recv(rtnetlink_socket, &req, sizeof(req), 0)) >= 0) {

			
		fprintf(stderr, "processing event from rtnetlink socket %d "
				" (%d)\n", rtnetlink_socket, recvsize);

		nlm = (struct nlmsghdr *) &req;
		while (NLMSG_OK(nlm, recvsize)) {
			switch (nlm->nlmsg_type) {
			case NLMSG_ERROR:
				fprintf(stderr,"error: %d (%s)\n",
				       ((struct nlmsgerr *)
					NLMSG_DATA(&req))->error,
				       strerror(-((struct nlmsgerr *)
						  NLMSG_DATA(&req))->
						error));
				exit(1);
				break;

			case RTM_NEWLINK:
				fprintf(stderr,"rtm_newlink ");

				rval = process_link_event(state,nlm);
				if(rval==state->ifindex){
					fprintf(stderr, "GOOD IFINDEX %d ",
							rval);

				}
				/*XXX: Bad hack to get testing going */
				/*else{
					fprintf(stderr, "NOT GOOD IFINDEX %d ",
							rval);
					rval=0;
				}*/
				fprintf(stderr,"\n");
				break;
			case RTM_DELLINK:
				fprintf(stderr,"rtm_dellink\n");

				return 0;
			case NLMSG_DONE:
				return rval;
				break;
			default:
				fprintf(stderr,"NLMSG Type: %d recvsize=%d\n",
				       req.nh.nlmsg_type, recvsize);
				break;
			}
			nlm = NLMSG_NEXT(nlm, recvsize);
		}
	}
	if ((recvsize < 0)&&(errno != EAGAIN)) {
		fprintf(stderr, "unable to receive  netlink message\n");
		perror("");
		exit(1);
	}
	return rval;
}


int get_all_events(int fd, struct link_status *state, struct prefixlist
			*linkplist, struct prefixlist *recvplist){
	int rval;
	int realevent=0;
	struct timeval now;
	struct tm *tm;

	while ((rval = get_li_rtnetlink(fd, state)) >= 0){
		if(!rval)
			break;
		realevent++;
		/*  0 return ifindex of link up??? */
		gettimeofday(&now, NULL);
		tm = localtime(&now.tv_sec);
		fprintf(stderr,"link up event: %d:%2.2d:%2.2d.%6.6lu\n",
		       tm->tm_hour, tm->tm_min, tm->tm_sec,
		       now.tv_usec);
	}
	if(realevent){
		state->new_hint_received=1;
		/* clear out the other fds */
		discard_ra(state->rasock, state, linkplist, recvplist);
		return state->ifindex;
	}
	return 0;
}



void timerrestore(struct link_status *state, struct prefixlist *linkplist,
				struct prefixlist *recvplist){
	state->tv.tv_sec=4*RTR_SOLICITATION_INTERVAL;

	printf("increment timers\n");

	if(state->hint_received || state->gather_prefixes){
 		if(state->rs_transmissions >= MAX_RTR_SOLICITATIONS){

			printf("Third RS tx complete\n");

			state->hint_received=0;
			state->gather_prefixes=0;
			union_prefixlist(linkplist,recvplist);
			linkplist->complete=1;
			print_prefixlist(linkplist);
			return;
		}

		printf("resend RS\n");

		state->tv.tv_sec=RTR_SOLICITATION_INTERVAL;
		state->rs_transmissions++;
		send_rs(state->v6sock, state->ifindex, state->macaddr,
			 state->tentative, &state->source, NULL,
			(uint8_t *)&state->landmark);
	}
}
