/* 2005/03/23 */

/* Configuration routines using (almost) pure RTNETLINK.*/

/* PURPOSE:

	This is a hacked example. Code style has been ignored

	Its primary uses are in my experiments and for
	illustration purposes.

*/

/*************************************************************/
/*
Please note that either this license MUST be displayed
or a particular instance of this code may be converted to 
the GNU General Public License v2 or greater.

Please see README.txt for details. 
*/

/*
Copyright (c) 2005, Monash University
All rights reserved.
Contributors: Greg Daley

Redistribution and use in source and binary forms, with or
without modification, are permitted provided that the following
conditions are met:

* Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

* Redistributions in binary form must reproduce the above
  copyright notice, this list of conditions and the following
  disclaimer in the documentation and/or other materials
  provided with the distribution.

* Neither the name of Monash University nor the names of its
  contributors may be used to endorse or promote products derived
  from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY
WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.

*/
/*************************************************************/




/*  GD: Note to self: 20040918 origin
    GD: Note to self: 20050110 evtest

     REF RFC 3549, for RTNETLINK msg  formats and types.


*/

/**************************************************************************/

/*ipv6 socket headers  AAA */
#include <sys/socket.h>
#include <netinet/in.h>
/* interface request headers  BBB */
#include <sys/ioctl.h>
#include <net/if.h>
/* extra ifreq headers*/
#include <string.h>

/* rtnetlink socket includes */
#include <asm/types.h>
#include <linux/netlink.h>
#include <linux/rtnetlink.h>
#include <sys/socket.h>


/* local diagnostic headers */
#include <stdio.h>
#include <stdlib.h>		/* exit(1)... */
#include <errno.h>

/* time information */
#include <time.h>
#include <sys/time.h>


/* handle wireless events */
#ifndef IFLA_WIRELESS
#define IFLA_WIRELESS 11	/* ???? */
#endif				/*IFLA_WIRELESS */

#ifndef _LINUX_IF_H
#define _LINUX_IF_H
#define _UNDEF_LINUX_IF_H
#endif				/*  _LINUX_IF_H */

#include <linux/wireless.h>

#ifdef _UNDEF_LINUX_IF_H
#undef _UNDEF_LINUX_IF_H
#undef _LINUX_IF_H
#endif				/* _UNDEF_LINUX_IF_H */

/* handle icmp6 filtering on reception */
#include <netinet/icmp6.h>

/* non blocking I/O on the ICMPv6 Socket */
#include <unistd.h>
#include <fcntl.h>

int bind_rtsock(uint32_t groups);
int open_v6sock(void);
int open_icmpv6receiver(void);
int get_ifindex(int sock, const char *name);

int set_rafilter(int sock);

int get_mac_rtnetlink(int sock, int ifindex, unsigned char *mac);
int get_linklocal_rtnetlink(int sock, int ifindex, unsigned char *lladdr);


/**************************************************************************/


int bind_rtsock(uint32_t groups)
{
	int rtnetlink_socket;
	struct sockaddr_nl local;

	memset(&local, 0, sizeof(local));
	local.nl_family = AF_NETLINK;
	local.nl_groups = groups;

	if ((rtnetlink_socket =
	     socket(PF_NETLINK, SOCK_DGRAM, NETLINK_ROUTE))
	    < 0) {
		fprintf(stderr, "get a netlink socket\n");
		exit(1);
	}

	if(fcntl(rtnetlink_socket, F_SETFL, O_NONBLOCK)<0){
		fprintf(stderr, "unable to get nonblock rtnl sock\n");
		exit(1);
	}

	if (bind
	    (rtnetlink_socket, (struct sockaddr *) &local,
	     sizeof(local)) < 0) {
		fprintf(stderr, "unable to bind netlink socket\n");
		exit(1);
	}
	printf("bound socket, prepare to send a netlink message\n");
	return rtnetlink_socket;
}

int open_v6sock(void)
{
	int sock;

	if ((sock = socket(AF_INET6, SOCK_RAW, IPPROTO_RAW)) < 0) {
		return -1;
	}
	return sock;
}

int open_icmpv6receiver(void)
{
	int info=1;
	int sock;

	if((sock = socket(AF_INET6, SOCK_RAW, IPPROTO_ICMPV6)) < 0){
		return -1;
	}

	if(fcntl(sock, F_SETFL, O_NONBLOCK)<0){
		return -1;
	}

	if(setsockopt(sock, IPPROTO_IPV6, IPV6_PKTINFO , &info, sizeof(int))<0){
		return -1;
	}
	
	return sock;
}

/* based on RFC3542 */

/* based on RFC3542 */

int set_rafilter(int sock){
        int ret;
	struct icmp6_filter f;

	ICMP6_FILTER_SETBLOCKALL( &f );
	ICMP6_FILTER_SETPASS( ND_ROUTER_ADVERT , &f );

	ret = setsockopt(sock, IPPROTO_ICMPV6, ICMP6_FILTER , &f, 
			sizeof(f));
	
	return ret;
}

int get_ifindex(int sock, const char *name)
{
	struct ifreq ifr;

	/* initialize the interface request structure */
	memset(&ifr, 0, sizeof(struct ifreq));

	/* get the name of the interface              */

	if (!strncpy(ifr.ifr_name, name, IFNAMSIZ)) {
		fprintf(stderr, "unable to set interface name %s\n", name);
		exit(1);
	}

	if (ioctl(sock, SIOCGIFINDEX, &ifr) < 0) {
		fprintf(stderr, "unable to get ifindex on fd: %d \n",
			sock);
		exit(1);
	}
	printf("received an ifindex for interface %s: %d\n",
	       ifr.ifr_name, ifr.ifr_ifindex);

	return ifr.ifr_ifindex;
}



/*------------------------------------------------------------*/

static int get_nlsequence(void)
{
	static int seq = 500;
	return ++seq;
}



int get_mac_rtnetlink(int rtnetlink_socket, int ifindex,
		      unsigned char *mac)
{
	struct rtattr *rta;
	unsigned char *rtap;	/* RTA_PAYLOAD */
	unsigned int rtal;
	int recvsize;
	struct {
		struct nlmsghdr nh;
		struct ifinfomsg ifi;
		char attrbuf[1024];
	} req;
	struct ifinfomsg *ifm;
	int ifl;
	struct nlmsghdr *nlm;
	int ret = 0;


	memset(&req, 0, sizeof(req));

	req.nh.nlmsg_len = NLMSG_LENGTH(sizeof(struct ifinfomsg));

	req.nh.nlmsg_flags = NLM_F_MATCH | NLM_F_REQUEST;
	req.nh.nlmsg_seq = get_nlsequence();
	req.nh.nlmsg_type = RTM_GETLINK;
	req.ifi.ifi_family = AF_INET6;
	req.ifi.ifi_index = ifindex;

	if (send(rtnetlink_socket, &req, req.nh.nlmsg_len, 0) < 0) {
		fprintf(stderr, "unable to send netlink message\n");
		exit(1);
	}

	printf("looking for MAC address (i:%d)\n",ifindex);
	while ((recvsize =
		recv(rtnetlink_socket, &req, sizeof(req),  0)) > 0) {

		nlm = (struct nlmsghdr *) &req;
		while (NLMSG_OK(nlm, recvsize)) {
			switch (nlm->nlmsg_type) {
			case NLMSG_ERROR:
				printf("error: %d (%s)\n",
				       ((struct nlmsgerr *)
					NLMSG_DATA(&req))->error,
				       strerror(-
						((struct nlmsgerr *)
						 NLMSG_DATA(&req))->
						error));
				exit(1);
				break;
			case RTM_NEWLINK:
				ifm = NLMSG_DATA(nlm);
				printf("Newlink with ifindex (%d)\n",
						ifm->ifi_index);
				if (ifindex != ifm->ifi_index) {
					/* skip things which aren't for us */
					break;
				}

				printf("Newlink with our ifindex (%d)\n",
						ifindex);
				ifl = nlm->nlmsg_len;
				rta = IFLA_RTA(ifm);

				while (RTA_OK(rta, ifl)) {
					switch (rta->rta_type) {
					case IFLA_ADDRESS:
						rtap = RTA_DATA(rta);
						rtal = RTA_PAYLOAD(rta);
						memcpy(mac, rtap, rtal);
						ret = ifindex;
						break;
					default:
						break;
					}
					rta = RTA_NEXT(rta, ifl);
				}
				break;
			case NLMSG_DONE:
				return ret;
				break;
			default:
				break;
			}
			nlm = NLMSG_NEXT(nlm, recvsize);
		}
	}
	if (recvsize < 0) {
		fprintf(stderr, "unable to receive  netlink message\n");
		perror("");
		exit(1);
	}
	return ret;
}



/*------------------------------------------------------------*/
int get_linklocal_rtnetlink(int rtnetlink_socket, int ifindex,
		      unsigned char *lladdr)
{
	struct rtattr *rta;
	unsigned char *rtap;	/* RTA_PAYLOAD */
	unsigned int rtal;
	int recvsize;
	struct {
		struct nlmsghdr nh;
		struct ifaddrmsg ia;
		char attrbuf[1024];
	} req;
	struct nlmsghdr *nlm;
	struct ifaddrmsg *iam;
	int ial;
	int ret = 0;


	memset(&req, 0, sizeof(req));

	req.nh.nlmsg_len = NLMSG_LENGTH(sizeof(struct ifinfomsg));

	req.nh.nlmsg_flags = NLM_F_MATCH | NLM_F_REQUEST;
	req.nh.nlmsg_seq = get_nlsequence();
	req.nh.nlmsg_type = RTM_GETADDR;
	req.ia.ifa_family = AF_INET6;
	req.ia.ifa_scope = RT_SCOPE_LINK;
	req.ia.ifa_index = ifindex;

	if (send(rtnetlink_socket, &req, req.nh.nlmsg_len, 0) < 0) {
		fprintf(stderr, "unable to send netlink message\n");
		exit(1);
	}

	while ((recvsize =
		recv(rtnetlink_socket, &req, sizeof(req), 0)) > 0) {


		nlm = (struct nlmsghdr *) &req;
		while (NLMSG_OK(nlm, recvsize)) {
			switch (nlm->nlmsg_type) {
			case NLMSG_ERROR:
				printf("error: %d (%s)\n",
				       ((struct nlmsgerr *)
					NLMSG_DATA(&req))->error,
				       strerror(-
						((struct nlmsgerr *)
						 NLMSG_DATA(&req))->
						error));
				exit(1);
				break;
			case RTM_NEWADDR:
				iam = NLMSG_DATA(nlm);
	
				if (ifindex != iam->ifa_index){
					/* skip things which aren't for us */
					break;
				}

				ial = nlm->nlmsg_len;
				rta = IFA_RTA(iam);

				while (RTA_OK(rta, ial)) {
					switch (rta->rta_type) {
					case IFA_ADDRESS:
						rtap = RTA_DATA(rta);
						rtal = RTA_PAYLOAD(rta);
						if((rtap[0] != 0xfe) ||
							(rtap[1] != 0x80)){
							break;
						}
						memcpy(lladdr, rtap, rtal);
						ret = ifindex;
						break;
					case IFA_LOCAL:
						rtap = RTA_DATA(rta);
						rtal = RTA_PAYLOAD(rta);
						memcpy(lladdr, rtap, rtal);
						ret = ifindex;
						break;
					default:
						break;
					}
					rta = RTA_NEXT(rta, ial);
				}
				break;
			case NLMSG_DONE:
				return ret;
				break;
			default:
				/*printf("RT Address message received: %d\n",
					nlm->nlmsg_type);*/
				break;
			}
			nlm = NLMSG_NEXT(nlm, recvsize);
		}
	}
	if (recvsize < 0) {
		fprintf(stderr, "unable to receive  netlink message\n");
		perror("");
		exit(1);
	}
	return ret;
}



/*------------------------------------------------------------*/

